use std::time::Instant;

use rand::{rngs::ThreadRng, seq::IteratorRandom, Rng};

use crate::graphics::Vertex;

use super::{map::CellKind, Map};

#[derive(Debug, Clone, Copy, PartialEq, Eq)]
pub enum EntityKind {
    Horse
}
impl EntityKind {
    const fn is_herbivore(&self) -> bool {
        match self {
            Self::Horse => true
        }
    }
}

#[derive(Debug, Clone, Copy, PartialEq, Eq)]
pub enum EntityState {
    Walking(Instant), // Start of walk
    Resting
}

#[derive(Debug)]
pub enum ExternOp {
    Remove,
    Move(usize)
}

#[derive(Debug)]
pub struct Entity {
    pub cid: usize,
    kind: EntityKind,
    start: Instant,
    state: EntityState,
    health: f32 // between 0 and 1
}
impl Entity {
    pub fn new(kind: EntityKind, cid: usize, ) -> Self {
        Self {
            cid,
            kind,
            start: Instant::now(),
            state: EntityState::Walking(Instant::now()),
            health: 1.
        }
    }
    pub fn set_health(&mut self, val: f32) {
        self.health = val.clamp(0., 1.);
    }
    pub fn render(&self, vertices: &mut Vec<Vertex>, indices: &mut Vec<u32>, map: &Map) {
        let pos = &map.voronoi.sites()[self.cid];
        match self.kind {
            EntityKind::Horse => {
                let color = [171./255. * self.health, 122./255. * self.health, 50./255. * self.health, 1.];
                let dark = [color[0]-0.3, color[1]-0.3, color[2]-0.3, 1.];
                let (vs, is) = match self.state {
                    EntityState::Walking(now) => {
                        let now = now.elapsed().as_secs_f32()*5.;
                        (
                            [
                                // back left leg
                                Vertex { pos: [-0.5, 0.3], color: dark },
                                Vertex { pos: [-0.4 + (now.sin()*0.1), -0.7 + (now.cos().max(-0.5)*0.1)], color: dark },
                                Vertex { pos: [-0.25, 0.1], color: dark },

                                // back right leg
                                Vertex { pos: [-0.5, 0.3], color },
                                Vertex { pos: [-0.4 + ((now + 1.).sin()*0.1), -0.7 + ((now + 1.).cos().max(-0.5)*0.1)], color },
                                Vertex { pos: [-0.25, 0.1], color },

                                // front left leg
                                Vertex { pos: [0.3, 0.2], color: dark },
                                Vertex { pos: [0.4 + ((now-1.).sin()*0.1), -0.7 + ((now-1.).cos().max(-0.5)*0.1)], color: dark },
                                Vertex { pos: [0.5, 0.3], color: dark },

                                // front right leg
                                Vertex { pos: [0.3, 0.2], color },
                                Vertex { pos: [0.4 + ((now-2.).sin()*0.1), -0.7 + ((now-2.).cos().max(-0.5)*0.1)], color },
                                Vertex { pos: [0.5, 0.3], color },

                                // body
                                // 3
                                Vertex { pos: [-0.3, 0.], color },
                                Vertex { pos: [0.4, -0.1], color },
                                // 11
                                Vertex { pos: [0.3, 0.4], color },
                            ],
                            [
                                0,1,2,
                                3,4,5,
                                6,7,8,
                                9,10,11,
                                3,12,13,
                                3,13,11,
                                3,11,14
                            ]
                        )
                    },
                    EntityState::Resting => {
                        (
                            [
                                // back left leg
                                Vertex { pos: [-0.5, 0.3], color: dark },
                                Vertex { pos: [-0.4, -0.75], color: dark },
                                Vertex { pos: [-0.25, 0.1], color: dark },

                                // back right leg
                                Vertex { pos: [-0.5, 0.3], color },
                                Vertex { pos: [-0.4, -0.75], color },
                                Vertex { pos: [-0.25, 0.1], color },

                                // front left leg
                                Vertex { pos: [0.3, 0.2], color: dark },
                                Vertex { pos: [0.4, -0.75], color: dark },
                                Vertex { pos: [0.5, 0.3], color: dark },

                                // front right leg
                                Vertex { pos: [0.3, 0.2], color },
                                Vertex { pos: [0.4, -0.75], color },
                                Vertex { pos: [0.5, 0.3], color },

                                // body
                                // 3
                                Vertex { pos: [-0.3, 0.], color },
                                Vertex { pos: [0.4, -0.1], color },
                                // 11
                                Vertex { pos: [0.3, 0.4], color },
                            ],
                            [
                                0,1,2,
                                3,4,5,
                                6,7,8,
                                9,10,11,
                                3,12,13,
                                3,13,11,
                                3,11,14
                            ]
                        )
                    }
                };

                vertices.reserve(vs.len());
                let base = vertices.len() as u32;
                for mut v in vs {
                    v.pos[0] = v.pos[0]/50. + pos.x as f32;
                    v.pos[1] = (v.pos[1] + 0.75)/50. + pos.y as f32;
                    vertices.push(v)
                }

                indices.reserve(is.len());
                for i in is {
                    indices.push(base + i);
                }
            }
        }
    }
    /// Returns new cell if entity moves
    pub fn update(&mut self, map: &mut Map, t: usize, rng: &mut ThreadRng) -> Option<ExternOp> {
        // Let’s take 0.57 kg of grass / m2
        // Let’s take 7.5 kg of food / day for a horse
        // Let’s say that a horse can survive up to 20 days without food
        let cd = &mut map.cells_data[self.cid];
        if self.kind.is_herbivore() {
            let food_needed: f32 = match self.kind {
                EntityKind::Horse => 7.5 / (0.57 * Map::CELL_AREA)
            }; // in cell resource fraction
            match cd.kind {
                CellKind::Forest => {}, // Infinite food in forests
                CellKind::Grass => {
                    let food_eaten = food_needed.min(cd.resource);
                    self.set_health(self.health - ((((food_needed - food_eaten)/food_needed)-0.5)/20.));
                    cd.set_resource(cd.resource-food_eaten, t);
                },
                _ => {
                    self.set_health(self.health - 1./20.);
                }
            }
        }

        if self.health == 0. {
            return Some(ExternOp::Remove);
        }

        let r = cd.resource;
        if cd.kind != CellKind::Grass {
            map.voronoi.cell(cd.cid).iter_neighbors().filter(|n| {
                let cd = &map.cells_data[*n];
                cd.kind == CellKind::Grass
            }).choose(rng).map(|c| ExternOp::Move(c))
        } else if r < 0.5 {
            let cd = &map.cells_data[self.cid];
            Some(ExternOp::Move(
                map.voronoi.cell(cd.cid)
                    .iter_neighbors()
                    .filter(|n| map.cells_data[*n].kind == CellKind::Grass)
                    .fold(cd, |acc, c| {
                        let cd = &map.cells_data[c];
                        if acc.kind != CellKind::Grass {
                            cd
                        } else if cd.kind != CellKind::Grass {
                            acc
                        } else if acc.resource > cd.resource {
                            acc
                        } else if acc.resource < cd.resource {
                            cd
                        } else if rng.gen_bool(0.5) {
                            cd
                        } else {
                            acc
                        }
                    }).cid
            ))
        } else {
            None
        }
    }
}